"""Bond Dissociation Energy (BDE) workflow."""

import itertools
from typing import Annotated, Any, Iterable, Self, TypeVar

from pydantic import AfterValidator, BaseModel, Field, PositiveInt, ValidationInfo, field_validator, model_validator

from ..base import round_optional_float
from ..mode import Mode
from ..molecule import Molecule
from ..types import UUID
from .multistage_opt import MultiStageOptMixin
from .workflow import Workflow

# the id of a mutable object may change, thus using object()
_sentinel_mso_mode = object()
_T = TypeVar("_T")


class BDE(BaseModel):
    """
    Bond Dissociation Energy (BDE) result.

    energy => (E_{fragment1} + E_{fragment2}) - E_{starting molecule}

    :param fragment_idxs: indices of the atoms in the fragment that was dissociated (1-indexed)
    :param energy: BDE in kcal/mol
    :param fragment_energies: energy of fragments
    :param calculations_uuids: calculation UUIDs
    """

    fragment_idxs: tuple[PositiveInt, ...]
    energy: Annotated[float | None, AfterValidator(round_optional_float(6))]
    fragment_energies: tuple[float | None, float | None]
    calculation_uuids: tuple[list[UUID | None], list[UUID | None]]

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        """
        Return a string representation of the BDE result.

        >>> BDE(fragment_idxs=(1, 2), energy=1.0, fragment_energies=(4, 2), calculation_uuids=([], []))
        <BDE (1, 2)  1.00>
        """
        energy = "None" if self.energy is None else f"{self.energy:>5.2f}"

        return f"<{type(self).__name__} {self.fragment_idxs} {energy}>"


class BDEWorkflow(Workflow, MultiStageOptMixin):
    """
    Bond Dissociation Energy (BDE) workflow.

    Uses the modes from MultiStageOptSettings to compute BDEs.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow
    :param multistage_opt_settings: set by mode unless mode=MANUAL (ignores additional settings if set)
    :param solvent: solvent to use for singlepoint
    :param xtb_preopt: pre-optimize with xtb (sets based on mode when None)
    :param frequencies: whether to calculate frequencies

    Overridden:
    :param mso_mode: Mode for MultiStageOptSettings

    Turned off:
    :param constraints: constraints to add (not supported)
    :param transition_state: whether this is a transition state (not supported)

    New:
    :param optimize_fragments: whether to optimize the fragments, or just the starting molecule (default depends on mode)
    :param atoms: atoms to dissociate (1-indexed)
    :param fragment_indices: fragments to dissociate (all fields feed into this, 1-indexed)
    :param all_CH: dissociate all C–H bonds
    :param all_CX: dissociate all C–X bonds (X ∈ {F, Cl, Br, I, At, Ts})
    :param optimization_calculation_uuids: UUIDs of initial optimization calculations
    :param optimization_energy: energy of optimized initial molecule
    :param bdes: BDE results
    """

    mso_mode: Mode = _sentinel_mso_mode  # type: ignore [assignment]
    optimize_fragments: bool = None  # type: ignore [assignment]

    atoms: tuple[PositiveInt, ...] = Field(default_factory=tuple)
    fragment_indices: tuple[tuple[PositiveInt, ...], ...] = Field(default_factory=tuple)

    all_CH: bool = False
    all_CX: bool = False

    # Results
    optimization_calculation_uuids: list[UUID | None] | None = None
    optimization_energy: float | None = None
    bdes: list[BDE] = Field(default_factory=list)

    def __str__(self) -> str:
        r"""
        Return a string representation of the BDE workflow.

        >>> print(BDEWorkflow(initial_molecule=Molecule.from_xyz("H 0 0 0\nF 0 0 1"), mode=Mode.METICULOUS, atoms=[1, 2]))
        BDEWorkflow METICULOUS
        (1,)
        (2,)
        """
        return f"{type(self).__name__} {self.mode.name}\n" + "\n".join(map(str, self.fragment_indices))

    @property
    def level_of_theory(self) -> str:
        return self.multistage_opt_settings.level_of_theory

    @property
    def energies(self) -> tuple[float | None, ...]:
        return tuple(bde.energy for bde in self.bdes)

    @field_validator("constraints", "transition_state")
    @classmethod
    def turned_off(cls, value: _T, info: ValidationInfo) -> _T:
        if value:
            raise ValueError(f"{info.field_name} not supported in BDE workflows.")

        return value

    @field_validator("initial_molecule", mode="before")
    @classmethod
    def no_charge_or_spin(cls, val: Molecule | dict[str, Any]) -> Molecule | dict[str, Any]:
        """Ensure the molecule has no charge or spin."""
        if isinstance(val, dict):
            mol = Molecule(**val)
        elif isinstance(val, Molecule):
            mol = val
        else:
            raise ValueError(f"{val=} is not a Molecule.")

        if mol.charge != 0 or mol.multiplicity != 1:
            raise ValueError("Charge and spin partitioning undefined for BDE, only neutral singlet molecules supported.")

        return val

    @model_validator(mode="before")
    @classmethod
    def set_mso_mode(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Set the MultiStageOptSettings mode to match current BDE mode."""
        values["mso_mode"] = values["mode"]
        return values

    @model_validator(mode="after")
    def validate_and_build(self) -> Self:
        """Validate atomic numbers and build the atoms field."""
        self.atoms = tuple(self.atoms)
        self.fragment_indices = tuple(map(tuple, self.fragment_indices))

        match self.mode:
            case Mode.RECKLESS:
                # GFN-FF doesn't support open-shell species
                self.optimize_fragments = False
            case Mode.RAPID | Mode.CAREFUL | Mode.METICULOUS:
                # Default on
                self.optimize_fragments = self.optimize_fragments or self.optimize_fragments is None
            case _:
                raise NotImplementedError(f"{self.mode} not implemented.")

        for atom in self.atoms:
            if atom > len(self.initial_molecule):
                raise ValueError(f"{atom=} is out of range.")
        for fragment_idxs in self.fragment_indices:
            if any(a > len(self.initial_molecule) for a in fragment_idxs):
                raise ValueError(f"{fragment_idxs=} contains atoms that are out of range.")

        if self.all_CH:
            self.atoms = self.atoms + tuple(H for _C, H in find_CH_bonds(self.initial_molecule))
        if self.all_CX:
            self.atoms = self.atoms + tuple(X for _C, X in find_CX_bonds(self.initial_molecule))

        # Combine atoms and fragments, remove duplicates, and sort
        self.fragment_indices = self.fragment_indices + tuple((a,) for a in self.atoms)
        assert isinstance(self.fragment_indices, tuple)
        self.fragment_indices = tuple(sorted(set(self.fragment_indices)))

        return self


def atomic_number_indices(molecule: Molecule, atomic_numbers: set[PositiveInt] | PositiveInt) -> tuple[PositiveInt, ...]:
    r"""
    Return the indices of the atoms with the given atomic numbers.

    :param molecule: Molecule of interest
    :param atomic_numbers: atomic number(s) of interest

    >>> H2O = Molecule.from_xyz("H 0 0 0\nO 0 0 1\nH 0 1 1")
    >>> atomic_number_indices(H2O, 1)
    (1, 3)
    >>> atomic_number_indices(H2O, {8, 1})
    (1, 2, 3)
    >>> atomic_number_indices(H2O, {6, 9})
    ()
    """
    if isinstance(atomic_numbers, int):
        return tuple(i for i, an in enumerate(molecule.atomic_numbers, start=1) if an == atomic_numbers)
    return tuple(i for i, an in enumerate(molecule.atomic_numbers, start=1) if an in atomic_numbers)


def find_CH_bonds(molecule: Molecule, distance_max: float = 1.2) -> Iterable[tuple[PositiveInt, PositiveInt]]:
    r"""
    Find all C–H bonds in the molecule.

    :param molecule: Molecule of interest
    :param distance_max: distance max for bond

    >>> H2O = Molecule.from_xyz("H 0 0 0\nO 0 0 1\nH 0 1 1")
    >>> CH4 = Molecule.from_xyz("C 0 0 0\nH 0 0 1\nH 0 1 0\nH 1 0 0\nH 0 0 -1")
    >>> ethane = Molecule.from_xyz("C -1 0 0\nC 1 0 0\nH -1 0 1\nH -1 0 -1\nH -1 1 0\nH 1 0 1\nH 1 0 -1\nH 1 1 0")
    >>> list(find_CH_bonds(H2O))
    []
    >>> list(find_CH_bonds(CH4))
    [(1, 2), (1, 3), (1, 4), (1, 5)]
    >>> list(find_CH_bonds(ethane))
    [(1, 3), (1, 4), (1, 5), (2, 6), (2, 7), (2, 8)]
    """
    yield from find_AB_bonds(molecule, 6, 1, distance_max)


def find_CX_bonds(molecule: Molecule) -> Iterable[tuple[PositiveInt, PositiveInt]]:
    r"""
    Find all C–X bonds in the molecule.

    :param molecule: Molecule of interest
    :param distance_max: distance max for bond

    >>> HCF = Molecule.from_xyz("H 0 0 0\nC 0 0 1\nF 0 1 1")
    >>> list(find_CX_bonds(HCF))
    [(2, 3)]
    """
    halogens = {9: 2.0, 17: 2.2, 35: 2.5, 53: 2.8, 85: 3.0, 117: 4.0}
    yield from itertools.chain.from_iterable(find_AB_bonds(molecule, 6, x, distance) for x, distance in halogens.items())


def find_AB_bonds(molecule: Molecule, a: int, b: int, distance_max: float) -> Iterable[tuple[PositiveInt, PositiveInt]]:
    r"""
    Find all A–B bonds in the molecule.

    :param molecule: Molecule of interest
    :param a: atomic number of atom A
    :param b: atomic number of atom B
    :param distance_max: distance max for bond

    >>> H2O = Molecule.from_xyz("H 0 0 0\nO 0 0 1\nH 0 1 1")
    >>> list(find_AB_bonds(H2O, 8, 1, 1.1))
    [(2, 1), (2, 3)]
    """

    def close(idxs: tuple[PositiveInt, PositiveInt]) -> bool:
        return molecule.distance(*idxs) < distance_max

    yield from filter(
        close,
        itertools.product(
            atomic_number_indices(molecule, a),
            atomic_number_indices(molecule, b),
        ),
    )



from typing import Optional

from .base import Base, LowercaseStrEnum
from .message import Message
from .molecule import Molecule
from .settings import Settings
from .status import Status
from .types import UUID


class StJamesVersion(LowercaseStrEnum):
    """
    Flag for which version of ``stjames`` this calculation was saved in.
    This will become meaningful later, when we have (probably) multiple versions.
    """

    V0 = "09072023"


class Calculation(Base):
    molecules: list[Molecule]

    settings: Settings = Settings()

    status: Status = Status.QUEUED

    name: Optional[str] = None
    elapsed: Optional[float] = None
    logfile: Optional[str] = None
    messages: list[Message] = []

    engine: Optional[str] = "peregrine"
    uuids: list[UUID | None] | None = None

    # not to be changed by end users, diff. versions will have diff. defaults
    json_format: str = StJamesVersion.V0



"""Docking workflow."""

from typing import Annotated

from pydantic import AfterValidator, ConfigDict, field_validator

from ..base import Base, round_float
from ..pdb import PDB
from ..types import UUID, Vector3D
from .workflow import Workflow


class Score(Base):
    """
    Pose with its score.

    :param pose: conformation of the ligand when docked
    :param score: score of the pose, in kcal/mol
    """

    pose: UUID | None  # for calculation
    score: Annotated[float, AfterValidator(round_float(3))]


class DockingWorkflow(Workflow):
    """
    Docking workflow.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param molecules: Molecules to dock (optional)
    :param smiles: SMILES strings of the ligands (optional)
    :param do_csearch: whether to csearch starting structures
    :param do_optimization: whether to optimize starting structures
    :param conformers: UUIDs of optimized conformers
    :param target: PDB of the protein
    :param pocket: center (x, y, z) and size (x, y, z) of the pocket

    Results:
    :param scores: docked poses sorted by score
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    do_csearch: bool = True
    do_optimization: bool = True
    conformers: list[UUID] = []

    target: PDB
    pocket: tuple[Vector3D, Vector3D]

    do_pose_hydrogen_refinement: bool = True
    scores: list[Score] = []

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        """Return a string representation of the Docking workflow."""
        desc = self.target.description
        target = desc.code or desc.title
        ligand = "".join(atom.atomic_symbol for atom in self.initial_molecule.atoms)

        return f"<{type(self).__name__} {target} {ligand}>"

    @field_validator("pocket", mode="after")
    def validate_pocket(cls, pocket: tuple[Vector3D, Vector3D]) -> tuple[Vector3D, Vector3D]:
        center, size = pocket
        if any(q <= 0 for q in size):
            raise ValueError(f"Pocket size must be positive, got: {size}")

        return pocket



from .base import Base, LowercaseStrEnum


class MessageType(LowercaseStrEnum):
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"


class Message(Base):
    type: MessageType
    title: str
    body: str



"""Contains functions for dealing with the .cif file format."""

import re
from collections import deque
from datetime import datetime
from typing import Any

import numpy as np

from .data import CODES


def mmcif_string_to_mmcif_dict(filestring: str) -> dict[str, Any]:
    """Takes a .cif filestring and turns into a ``dict`` which represents its
    table structure. Only lines which aren't empty and which don't begin with
    ``#`` are used.

    Multi-line strings are consolidated onto one line, and the whole thing is
    then split into the blocks that will become table lists. At the end, quote
    marks are removed from any string which retains them.

    :param str filestring: the .cif filestring to process.
    :rtype: ``dict``"""

    lines = deque(filter(lambda l: l and l[0] != "#", filestring.split("\n")))
    lines = consolidate_strings(lines)
    blocks = mmcif_lines_to_mmcif_blocks(lines)
    mmcif_dict = {}
    for block in blocks:
        if block["lines"][0] == "loop_":
            mmcif_dict[block["category"]] = loop_block_to_list(block)
        else:
            mmcif_dict[block["category"]] = non_loop_block_to_list(block)
    strip_quotes(mmcif_dict)
    return mmcif_dict


def consolidate_strings(lines: deque[str]) -> deque[str]:
    """Generally, .cif files have a one file line to one table row
    correspondence. Sometimes however, a string cell is given a line of its own,
    breaking the row over several lines. This function takes the lines of a .cif
    file and puts all table rows on a single line.

    :param deque lines: the .cif file lines.
    :rtype: ``deque``"""

    new_lines: deque[str] = deque()
    while lines:
        line = lines.popleft()
        if line.startswith(";"):
            string = [line[1:].strip()]
            while not lines[0].startswith(";"):
                string.append(lines.popleft())
            lines.popleft()
            new_lines[-1] += ' "{}"'.format(" ".join(string).replace('"', "\x1a").replace("'", "\x1b"))
        else:
            new_lines.append(line)
    return new_lines


def mmcif_lines_to_mmcif_blocks(lines: deque[str]) -> list[dict[str, Any]]:
    """A .cif file is ultimately a list of tables. This function takes a list of
    .cif file lines and splits them into these table blocks. Each block will be
    a ``dict`` containing a category name and a list of lines.

    :param deque lines: the .cif file lines.
    :rtype: ``list``"""

    category = None
    block: list[str] = []
    blocks = []
    while lines:
        line = lines.popleft()
        if line.startswith("data_"):
            continue
        if line.startswith("_"):
            line_category = line.split(".")[0]
            if line_category != category:
                if category:
                    blocks.append({"category": category[1:], "lines": block})
                category = line_category
                block = []
        if line.startswith("loop_"):
            if category:
                blocks.append({"category": category[1:], "lines": block})
            category = lines[0].split(".")[0]
            block = []
        block.append(line)
    if block:
        blocks.append({"category": category[1:], "lines": block})  # type: ignore [index]
    return blocks


def non_loop_block_to_list(block: dict[str, Any]) -> list[dict[str, Any]]:
    """Takes a simple block ``dict`` with no loop and turns it into a table
    ``list``.

    :param dict block: the .cif block to process.
    :rtype: ``list``"""

    d = {}
    for index in range(len(block["lines"]) - 1):
        if block["lines"][index + 1][0] != "_":
            block["lines"][index] += " " + block["lines"][index + 1]
    block["lines"] = [l for l in block["lines"] if l[0] == "_"]
    for line in block["lines"]:
        name = line.split(".")[1].split()[0]
        value = line
        if line.startswith("_"):
            value = " ".join(line.split()[1:])
        d[name] = value
    return [d]


def loop_block_to_list(block: dict[str, Any]) -> list[dict[str, Any]]:
    """Takes a loop block ``dict`` where the initial lines are table headers and
    turns it into a table ``list``. Sometimes a row is broken over several lines
    so this function deals with that too.

    :param dict block: the .cif block to process.
    :rtype: ``list``"""

    names, lines, _ = [], [], True
    body_start = 0
    for index, line in enumerate(block["lines"][1:], start=1):
        if not line.startswith("_" + block["category"]):
            body_start = index
            break
    names = [l.split(".")[1].rstrip() for l in block["lines"][1:body_start]]
    lines = [split_values(l) for l in block["lines"][body_start:]]
    l = []
    for n in range(len(lines) - 1):
        while n < len(lines) - 1 and len(lines[n]) + len(lines[n + 1]) <= len(names):
            lines[n] += lines[n + 1]
            lines.pop(n + 1)
    for line in lines:
        l.append({name: value for name, value in zip(names, line)})
    return l


def split_values(line: str) -> list[str]:
    """The body of a .cif table is a series of lines, with each cell divided by
    whitespace. This function takes a string line and breaks it into cells.

    There are a few peculiarities to handle. Sometimes a cell is a string
    enclosed in quote marks, and spaces within this string obviously shouldn't
    be used to break the line. This function handles all of that.

    :param str line: the .cif line to split.
    :rtype: ``list``"""

    if not re.search("['\"]", line):
        return line.split()
    chars = deque(line.strip())
    values, in_string = [], False
    value: list[Any] = []
    while chars:
        char = chars.popleft()
        if char == " " and not in_string:
            values.append(value)
            value = []
        elif char in "'\"":
            if in_string and chars and chars[0] != " ":
                value.append(char)
            else:
                in_string = not in_string
        else:
            value.append(char)
    values.append(value)
    return ["".join(v) for v in values if v]


def strip_quotes(mmcif_dict: dict[str, Any]) -> None:
    """Goes through each table in the mmcif ``dict`` and removes any unneeded
    quote marks from the cells.

    :param dict mmcif_dict: the almost finished .mmcif dictionary to clean."""

    for name, table in mmcif_dict.items():
        for row in table:
            for k, value in row.items():
                for char in "'\"":
                    if value[0] == char and value[-1] == char:
                        row[k] = value[1:-1]
                    row[k] = row[k].replace("\x1a", '"').replace("\x1b", "'")


def mmcif_dict_to_data_dict(mmcif_dict: dict[str, Any]) -> dict[str, Any]:
    """Converts an .mmcif dictionary into an atomium data dictionary, with the
    same standard layout that the other file formats get converted into.

    :param dict mmcif_dict: the .mmcif dictionary.
    :rtype: ``dict``"""

    data_dict = {
        "description": {"code": None, "title": None, "deposition_date": None, "classification": None, "keywords": [], "authors": []},
        "experiment": {"technique": None, "source_organism": None, "expression_system": None, "missing_residues": []},
        "quality": {"resolution": None, "rvalue": None, "rfree": None},
        "geometry": {"assemblies": [], "crystallography": {}},
        "models": [],
    }
    update_description_dict(mmcif_dict, data_dict)
    update_experiment_dict(mmcif_dict, data_dict)
    update_quality_dict(mmcif_dict, data_dict)
    update_geometry_dict(mmcif_dict, data_dict)
    update_models_list(mmcif_dict, data_dict)
    return data_dict


def update_description_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """Takes a data dictionary and updates its description sub-dictionary with
    information from a .mmcif dictionary.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update."""

    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "code", "entry", "id")
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "title", "struct", "title")
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "deposition_date", "pdbx_database_status", "recvd_initial_deposition_date", date=True)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "classification", "struct_keywords", "pdbx_keywords")
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "keywords", "struct_keywords", "text", split=True)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "authors", "audit_author", "name", multi=True)


def update_experiment_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """Takes a data dictionary and updates its experiment sub-dictionary with
    information from a .mmcif dictionary.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update."""

    mmcif_to_data_transfer(mmcif_dict, data_dict, "experiment", "technique", "exptl", "method")
    for cat, key in [
        ["entity_src_nat", "pdbx_organism_scientific"],
        ["entity_src_gen", "pdbx_gene_src_scientific_name"],
        ["pdbx_entity_src_syn", "organism_scientific"],
    ]:
        mmcif_to_data_transfer(mmcif_dict, data_dict, "experiment", "source_organism", cat, key)
        if data_dict["experiment"]["source_organism"] not in [None, "?"]:
            break
    mmcif_to_data_transfer(mmcif_dict, data_dict, "experiment", "expression_system", "entity_src_gen", "pdbx_host_org_scientific_name")
    for r in mmcif_dict.get("pdbx_unobs_or_zero_occ_residues", []):
        insert = "" if r["PDB_ins_code"] in "?." else r["PDB_ins_code"]
        data_dict["experiment"]["missing_residues"].append({"id": f"{r['auth_asym_id']}.{r['auth_seq_id']}{insert}", "name": r["auth_comp_id"]})


def update_quality_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """Takes a data dictionary and updates its quality sub-dictionary with
    information from a .mmcif dictionary.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update."""

    mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "resolution", "reflns", "d_resolution_high", func=float)
    if not data_dict["quality"]["resolution"]:
        mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "resolution", "refine", "ls_d_res_high", func=float)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "rvalue", "refine", "ls_R_factor_R_work", func=float)
    if not data_dict["quality"]["rvalue"]:
        mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "rvalue", "refine", "ls_R_factor_obs", func=float)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "rfree", "refine", "ls_R_factor_R_free", func=float)


def update_geometry_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """Takes a data dictionary and updates its geometry sub-dictionary with
    information from a .mmcif dictionary.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update."""

    data_dict["geometry"]["assemblies"] = [
        {
            "id": int(a["id"]),
            "software": a.get("method_details", None),
            "delta_energy": None,
            "buried_surface_area": None,
            "surface_area": None,
            "transformations": [],
        }
        for a in mmcif_dict.get("pdbx_struct_assembly", [])
    ]
    operations = {
        o["id"]: [[float(o["matrix[{}][{}]".format(r, c)]) for c in [1, 2, 3]] + [float(o["vector[{}]".format(r)])] for r in [1, 2, 3]] + [[0, 0, 0, 1]]
        for o in mmcif_dict.get("pdbx_struct_oper_list", [])
    }
    for assembly in data_dict["geometry"]["assemblies"]:
        if assembly["software"] == "?":
            assembly["software"] = None
        assign_metrics_to_assembly(mmcif_dict, assembly)
        assign_transformations_to_assembly(mmcif_dict, operations, assembly)
    update_crystallography_dict(mmcif_dict, data_dict)


def assign_metrics_to_assembly(mmcif_dict: dict[str, Any], assembly: dict[str, Any]) -> None:
    """Takes an assembly dict, and goes through an mmcif dictionary looking for
    relevant energy etc. information to update it with.

    :param dict mmcif_dict: The dictionary to read.
    :param dict assembly: The assembly to update."""

    for a in mmcif_dict.get("pdbx_struct_assembly_prop", []):
        if a["biol_id"] == str(assembly["id"]):
            if a["type"] == "MORE":
                assembly["delta_energy"] = float(a["value"].split("/")[0])
            elif a["type"] == "SSA (A^2)":
                assembly["surface_area"] = float(a["value"].split("/")[0])
            elif a["type"] == "ABSA (A^2)":
                assembly["buried_surface_area"] = float(a["value"].split("/")[0])


def assign_transformations_to_assembly(mmcif_dict: dict[str, Any], operations: Any, assembly: dict[str, Any]) -> None:
    """Takes an assembly dict, and goes through an mmcif dictionary looking for
    relevant transformation information to update it with.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict operations: the processed operations matrices.
    :param dict assembly: the assembly to update."""

    for gen in mmcif_dict.get("pdbx_struct_assembly_gen", []):
        if gen["assembly_id"] == str(assembly["id"]):
            op_ids_groups = get_operation_id_groups(gen["oper_expression"])
            ops = operation_id_groups_to_operations(operations, op_ids_groups)
            for operation in ops:
                assembly["transformations"].append(
                    {"chains": gen["asym_id_list"].split(","), "matrix": [row[:3] for row in operation[:3]], "vector": [row[-1] for row in operation[:3]]}
                )


def get_operation_id_groups(expression: str) -> list[list[str]]:
    """Takes an operator expression from an .mmcif transformation dict, and
    works out what transformation IDs it is referring to. For example, (1,2,3)
    becomes [[1, 2, 3]], (1-3)(8-11,17) becomes [[1, 2, 3], [8, 9, 10, 11, 17]],
    and so on.

    :param str expression: The expression to parse.
    :rtype: ``list``"""

    if expression[0] != "(":
        expression = "({})".format(expression)
    groups = re.findall(r"\((.+?)\)", expression)
    group_ids = []
    for group in groups:
        ids = []
        elements = group.split(",")
        for element in elements:
            if "-" in element:
                bounds = [int(x) for x in element.split("-")]
                ids += [str(n) for n in list(range(bounds[0], bounds[1] + 1))]
            else:
                ids.append(element)
        group_ids.append(ids)
    return group_ids


def update_crystallography_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """Takes a data dictionary and updates its crystallography
    sub-sub-dictionary with information from a .mmcif dictionary.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update."""

    if mmcif_dict.get("cell"):
        mmcif_to_data_transfer(mmcif_dict, data_dict["geometry"], "crystallography", "space_group", "symmetry", "space_group_name_H-M")
        data_dict["geometry"]["crystallography"]["unit_cell"] = [
            float(mmcif_dict["cell"][0][key].replace("?", "0")) for key in ["length_a", "length_b", "length_c", "angle_alpha", "angle_beta", "angle_gamma"]
        ]
    if data_dict["geometry"]["crystallography"].get("space_group") == "NA":
        data_dict["geometry"]["crystallography"] = {}


def operation_id_groups_to_operations(operations: Any, operation_id_groups: Any) -> Any:
    """Creates a list of operation matrices for an assembly, from a list of
    operation IDs - cross multiplying as required.

    :param dict operations: the parsed .mmcif operations.
    :param list operation_id_groups: the operation IDs."""

    operation_groups = [[operations[i] for i in ids] for ids in operation_id_groups]
    while len(operation_groups) and len(operation_groups) != 1:
        operations = []
        for op1 in operation_groups[0]:
            for op2 in operation_groups[1]:
                operations.append(np.matmul(op1, op2))
        operation_groups[0] = operations
        operation_groups.pop(1)
    return operation_groups[0]


def update_models_list(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """Takes a data dictionary and updates its models list with
    information from a .mmcif dictionary.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update."""

    data_dict["models"] = []
    types = {e["id"]: e["type"] for e in mmcif_dict.get("entity", {})}
    names = {e["id"]: e["name"] for e in mmcif_dict.get("chem_comp", {}) if e["mon_nstd_flag"] != "y"}
    entities = {m["id"]: m["entity_id"] for m in mmcif_dict.get("struct_asym", [])}
    # sequences = make_sequences(mmcif_dict)
    secondary_structure = make_secondary_structure(mmcif_dict)
    aniso = make_aniso(mmcif_dict)
    model: dict[str, Any] = {"polymer": {}, "non_polymer": {}, "water": {}, "branched": {}}
    model_num = mmcif_dict["atom_site"][0]["pdbx_PDB_model_num"]
    for atom in mmcif_dict["atom_site"]:
        if atom["pdbx_PDB_model_num"] != model_num:
            data_dict["models"].append(model)
            model = {"polymer": {}, "non_polymer": {}, "water": {}, "branched": {}}
            model_num = atom["pdbx_PDB_model_num"]
        mol_type = types[entities[atom["label_asym_id"]]]
        if mol_type == "polymer" or mol_type == "branched":
            add_atom_to_polymer(atom, aniso, model, names)
        else:
            add_atom_to_non_polymer(atom, aniso, model, mol_type, names)
    data_dict["models"].append(model)
    for model in data_dict["models"]:
        add_sequences_to_polymers(model, mmcif_dict, entities)
        add_secondary_structure_to_polymers(model, secondary_structure)


def make_aniso(mmcif_dict: dict[str, Any]) -> dict[int, Any]:
    """Makes a mapping of atom IDs to anisotropy information.

    :param mmcif_dict: the .mmcif dict to read.
    :rtype: ``dict``"""

    return {
        int(a["id"]): [float(a["U[{}][{}]".format(x, y)]) for x, y in ["11", "22", "33", "12", "13", "23"]]  # type: ignore [has-type, misc]
        for a in mmcif_dict.get("atom_site_anisotrop", [])
    }


def make_secondary_structure(mmcif_dict: dict[str, Any]) -> dict[str, Any]:
    """Creates a dictionary of helices and strands, with each having a list of
    start and end residues.

    :param mmcif_dict: the .mmcif dict to read.
    :rtype: ``dict``"""

    helices, strands = [], []
    for helix in mmcif_dict.get("struct_conf", []):
        helices.append(
            [
                "{}.{}{}".format(
                    helix[f"{x}_auth_asym_id"],
                    helix[f"{x}_auth_seq_id"],
                    helix[f"pdbx_{x}_PDB_ins_code"].replace("?", ""),
                )
                for x in ["beg", "end"]
            ]
        )
    for strand in mmcif_dict.get("struct_sheet_range", []):
        strands.append(
            [
                "{}.{}{}".format(
                    strand[f"{x}_auth_asym_id"],
                    strand[f"{x}_auth_seq_id"],
                    strand[f"pdbx_{x}_PDB_ins_code"].replace("?", ""),
                )
                for x in ["beg", "end"]
            ]
        )
    return {"helices": helices, "strands": strands}


def add_atom_to_polymer(atom: dict[str, Any], aniso: dict[int, Any], model: dict[str, Any], names: dict[str, Any]) -> None:
    """Takes an MMCIF atom dictionary, converts it, and adds it to a polymer
    dictionary.

    :param dict atom: the .mmcif dictionary to read.
    :param dict aniso: lookup dictionary for anisotropy information.
    :param dict model: the model to update.
    :param dict names: the lookup dictionary for full name information."""

    mol_id = atom["auth_asym_id"]
    res_id = make_residue_id(atom)
    try:
        model["polymer"][mol_id]["residues"][res_id]["atoms"][int(atom["id"])] = atom_dict_to_atom_dict(atom, aniso)
    except Exception:
        name = atom["auth_comp_id"]
        try:
            model["polymer"][mol_id]["residues"][res_id] = {
                "name": name,
                "full_name": names.get(name),
                "atoms": {int(atom["id"]): atom_dict_to_atom_dict(atom, aniso)},
                "number": len(model["polymer"][mol_id]["residues"]) + 1,
            }
        except Exception:
            model["polymer"][mol_id] = {
                "internal_id": atom["label_asym_id"],
                "helices": [],
                "strands": [],
                "residues": {
                    res_id: {
                        "name": name,
                        "atoms": {int(atom["id"]): atom_dict_to_atom_dict(atom, aniso)},
                        "number": 1,
                        "full_name": names.get(name),
                    }
                },
            }


def add_atom_to_non_polymer(atom: dict[str, Any], aniso: dict[int, Any], model: dict[str, Any], mol_type: str, names: dict[str, Any]) -> None:
    """Takes an MMCIF atom dictionary, converts it, and adds it to a non_polymer
    dictionary.

    :param dict atom: the .mmcif dictionary to read.
    :param dict aniso: lookup dictionary for anisotropy information.
    :param dict model: the model to update.
    :param str mol_type: non_polymer or water.
    :param dict names: the lookup dictionary for full name information."""

    mol_id = make_residue_id(atom)
    try:
        model[mol_type][mol_id]["atoms"][int(atom["id"])] = atom_dict_to_atom_dict(atom, aniso)
    except Exception:
        name = atom["auth_comp_id"]
        model[mol_type][mol_id] = {
            "name": name,
            "full_name": names.get(name),
            "internal_id": atom["label_asym_id"],
            "polymer": atom["auth_asym_id"],
            "atoms": {int(atom["id"]): atom_dict_to_atom_dict(atom, aniso)},
        }


def make_residue_id(d: dict[str, Any]) -> str:
    """Generates a residue ID for an atom.

    :param dict d: the atom dictionary to read.
    :rtype: ``str``"""

    insert = "" if d["pdbx_PDB_ins_code"] in "?." else d["pdbx_PDB_ins_code"]
    return "{}.{}{}".format(d["auth_asym_id"], d["auth_seq_id"], insert)


def add_sequences_to_polymers(model: dict[str, Any], mmcif_dict: dict[str, Any], entities: dict[str, Any]) -> None:
    """Takes a pre-populated mapping of chain IDs to entity IDs, and uses them
    to add sequence information to a model.

    :param dict model: the model to update.
    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict entities: a mapping of chain IDs to entity IDs."""

    sequences = make_sequences(mmcif_dict)
    for polymer in model["polymer"].values():
        polymer["sequence"] = sequences.get(entities.get(polymer["internal_id"], ""), "")


def add_secondary_structure_to_polymers(model: dict[str, Any], ss_dict: dict[str, Any]) -> None:
    """Updates polymer dictionaries with secondary structure information, from
    a previously created mapping.

    :param dict model: the model to update.
    :param dict ss_dict: the mapping to read."""

    for ss in ("helices", "strands"):
        for segment in ss_dict[ss]:
            chain = model["polymer"].get(segment[0].split(".")[0])
            if chain:
                in_segment = False
                chain[ss].append([])
                for residue_id in chain["residues"].keys():
                    if residue_id == segment[0]:
                        in_segment = True
                    if in_segment:
                        chain[ss][-1].append(residue_id)
                    if residue_id == segment[1]:
                        break


def make_sequences(mmcif_dict: dict[str, Any]) -> dict[str, Any]:
    """Creates a mapping of entity IDs to sequences.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :rtype: ``dict``"""

    return {
        e["id"]: "".join([CODES.get(res["mon_id"], "X") for res in mmcif_dict.get("entity_poly_seq", []) if res["entity_id"] == e["id"]])
        for e in mmcif_dict.get("entity", [])
        if e["type"] == "polymer"
    }


def atom_dict_to_atom_dict(d: dict[str, Any], aniso_dict: dict[int, Any]) -> dict[str, Any]:
    """Turns an .mmcif atom dictionary into an atomium atom data dictionary.

    :param dict d: the .mmcif atom dictionary.
    :param dict d: the mapping of atom IDs to anisotropy.
    :rtype: ``dict``"""

    charge = "pdbx_formal_charge"
    atom = {
        "x": d["Cartn_x"],
        "y": d["Cartn_y"],
        "z": d["Cartn_z"],
        "element": d["type_symbol"],
        "name": d.get("label_atom_id"),
        "occupancy": d.get("occupancy", 1),
        "bvalue": d.get("B_iso_or_equiv"),
        "charge": d.get(charge, 0) if d.get(charge) != "?" else 0,
        "alt_loc": d.get("label_alt_id") if d.get("label_alt_id") != "." else None,
        "anisotropy": aniso_dict.get(int(d["id"]), [0, 0, 0, 0, 0, 0]),
        "is_hetatm": d.get("group_PDB", "ATOM") == "HETATM",
    }
    for key in ["x", "y", "z", "charge", "bvalue", "occupancy"]:
        if atom[key] is not None:
            atom[key] = float(atom[key])
    return atom


def mmcif_to_data_transfer(
    mmcif_dict: dict[str, Any],
    data_dict: dict[str, Any],
    d_cat: Any,
    d_key: Any,
    m_table: Any,
    m_key: Any,
    date: bool = False,
    split: bool = False,
    multi: bool = False,
    func: Any = None,
) -> None:
    """A function for transfering a bit of data from a .mmcif dictionary to a
    data dictionary, or doing nothing if the data doesn't exist.

    :param dict mmcif_dict: the .mmcif dictionary to read.
    :param dict data_dict: the data dictionary to update.
    :param str d_cat: the top-level key in the data dictionary.
    :param str d_key: the data dictionary field to update.
    :param str m_table: the name of the .mmcif table to look in.
    :param str m_key: the .mmcif field to read.
    :param bool date: if True, the value will be converted to a date.
    :param bool split: if True, the value will be split on commas.
    :param bool multi: if True, every row in the table will be read.
    :param function func: if given, this will be applied to the value."""

    try:
        if multi:
            value = [row[m_key] for row in mmcif_dict[m_table]]
        else:
            value = mmcif_dict[m_table][0][m_key]
        if date:
            value = datetime.strptime(value, "%Y-%m-%d").date()  # type: ignore [arg-type, assignment]
        if split:
            value = value.replace(", ", ",").split(",")  # type: ignore [attr-defined]
        if func:
            value = func(value)
        data_dict[d_cat][d_key] = None if value == "?" else value  # type: ignore [comparison-overlap]
    except Exception:
        pass



from ..settings import Settings
from ..types import UUID
from .workflow import Workflow


class BasicCalculationWorkflow(Workflow):
    settings: Settings
    engine: str
    calculation_uuid: UUID | None = None



from typing import Annotated, Any

from pydantic import AfterValidator, BaseModel, Field, PositiveInt, field_validator, model_validator

from ..base import round_float
from ..mode import Mode
from ..types import UUID
from .multistage_opt import MultiStageOptMixin
from .workflow import Workflow


class SpinState(BaseModel):
    """
    The result of a SpinState calculation.

    :param multiplicity: multiplicity of the SpinState
    :param energy: energy of the optimized Molecule
    :param calculation: the UUIDs of the Calculations that produced this SpinState

    >>> from stjames.molecule import Atom, Molecule
    >>> He = Molecule(charge=0, multiplicity=1, atoms=[Atom(atomic_number=2, position=[0, 0, 0])])
    >>> SpinState(multiplicity=3, energy=-13.4291, calculation=['8a031a27-30d2-4ac7-8ade-efae9e9fc94a'])
    <SpinState 3 -13.429>
    """

    multiplicity: PositiveInt
    energy: Annotated[float, AfterValidator(round_float(6))]
    calculation: list[UUID | None]

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.multiplicity} {self.energy:.3f}>"


_sentinel_mso_mode = object()


class SpinStatesWorkflow(Workflow, MultiStageOptMixin):
    """
    Workflow for computing spin states of molecules.

    Uses the modes from MultiStageOptSettings.

    Influenced by:
    [Performance of Quantum Chemistry Methods for Benchmark Set of Spin–State
    Energetics Derived from Experimental Data of 17 Transition Metal Complexes
    (SSE17)](https://chemrxiv.org/engage/chemrxiv/article-details/66a8b15cc9c6a5c07a792487)

    Inherited
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow
    :param multistage_opt_settings: set by mode unless mode=MANUAL (ignores additional settings if set)
    :param solvent: solvent to use for optimization
    :param xtb_preopt: pre-optimize with xtb
    :param constraints: constraints to add
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    Overridden:
    :param mso_mode: Mode for MultiStageOptSettings

    New:
    :param states: multiplicities of the spin state targetted
    :param spin_states: resulting spin states data


    >>> from stjames.molecule import Atom, Molecule
    >>> He = Molecule(charge=0, multiplicity=1, atoms=[Atom(atomic_number=2, position=[0, 0, 0])])
    >>> ss = SpinStatesWorkflow(initial_molecule=He, states=[1, 3, 5], mode=Mode.RAPID, solvent="water")
    >>> str(ss)
    '<SpinStatesWorkflow [1, 3, 5] RAPID>'
    """

    mso_mode: Mode = _sentinel_mso_mode  # type: ignore [assignment]
    states: list[PositiveInt]

    # Results
    spin_states: list[SpinState] = Field(default_factory=list)

    def __repr__(self) -> str:
        if self.mode != Mode.MANUAL:
            return f"<{type(self).__name__} {self.states} {self.mode.name}>"

        return f"<{type(self).__name__} {self.states} {self.level_of_theory}>"

    def __len__(self) -> int:
        return len(self.states)

    @property
    def level_of_theory(self) -> str:
        return self.multistage_opt_settings.level_of_theory

    @field_validator("states")
    @classmethod
    def validate_states(cls, states: list[PositiveInt]) -> list[PositiveInt]:
        """Confirm that all spin states are valid."""
        if not states:
            raise ValueError("Expected at least one spin state.")

        if any((states[0] - mult) % 2 for mult in states):
            raise ValueError(f"Inconsistent multiplicities found: {states}")

        return states

    def str_results(self) -> str:
        return "SpinStatesResults\n" + "\n".join(f"  {ss.multiplicity}: {ss.energy:>5.2f}" for ss in self.spin_states)

    @property
    def energies(self) -> list[float]:
        return [ss.energy for ss in self.spin_states]

    @model_validator(mode="before")
    @classmethod
    def set_mso_mode(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Set the MultiStageOptSettings mode to match current SpinStates mode."""
        values["mso_mode"] = values["mode"]
        return values

    @field_validator("spin_states")
    @classmethod
    def validate_spin_states(cls, spin_states: list[SpinState]) -> list[SpinState]:
        """Ensure that all spin states have consistent results."""
        mults = [ss.multiplicity for ss in spin_states]

        if len(mults) != len(set(mults)):
            raise ValueError(f"Duplicate multiplicities found: {mults}")

        if any((mults[0] - mult) % 2 for mult in mults[1:]):
            raise ValueError(f"Inconsistent multiplicities found: {mults}")

        return spin_states



from typing import Annotated

from pydantic import AfterValidator, NonNegativeFloat, NonNegativeInt

from ..base import Base, round_float
from ..settings import Settings
from ..types import UUID, FloatPerAtom, Matrix3x3, Vector3D, round_optional_float_per_atom, round_optional_matrix3x3, round_optional_vector3d
from .workflow import Workflow


class PropertyCubePoint(Base):
    """A point in a cube file, all values rounded to 6 decimal places."""

    x: Annotated[float, AfterValidator(round_float(3))]
    y: Annotated[float, AfterValidator(round_float(3))]
    z: Annotated[float, AfterValidator(round_float(3))]
    val: Annotated[float, AfterValidator(round_float(6))]


class PropertyCube(Base):
    """Represents a "cubefile" of some property."""

    cube_points: list[PropertyCubePoint]


class MolecularOrbitalCube(PropertyCube):
    """
    Cube of a molecular orbital.

    Inherits `cube_data`.
    """

    occupation: NonNegativeInt
    energy: Annotated[float, AfterValidator(round_float(6))]


class ElectronicPropertiesWorkflow(Workflow):
    """
    Workflow for computing electronic properties.

    Inherited
    :param initial_molecule: Molecule of interest

    Config settings:
    :param settings: settings for the calculation
    :param compute_density_cube: whether to compute the density cube
    :param compute_electrostatic_potential_cube: whether to compute the electrostatic potential cube
    :param compute_num_occupied_orbitals: number of occupied orbitals to save
    :param compute_num_virtual_orbitals: number of virtual orbitals to save

    Populated while running:
    :param calc_uuid: UUID of the calculation
    :param dipole: dipole moment
    :param quadrupole: quadrupole moment
    :param lowdin_charges: Löwdin charges
    :param mulliken_charges: Mulliken charges
    :param wiberg_bond_orders: Wiberg bond orders (`atom1`, `atom2`, `order`)
    :param mayer_bond_orders: Mayer bond orders (`atom1`, `atom2`, `order`)

    :param density_cube: electron density, as a cube
    :param density_cube_alpha: α electron density, as a cube
    :param density_cube_beta: β electron density, as a cube
    :param density_cube_difference: difference spin densities, as a cube

    :param electrostatic_potential_cube: electrostatic potential, as a cube

    :param molecular_orbitals: MOs, key is absolute orbital index (for closed-shell species (RHF))
    :param molecular_orbitals_alpha: α MOs, key is absolute orbital index (for open-shell species (UHF/ROHF))
    :param molecular_orbitals_beta: β MOs, key is absolute orbital index (for open-shell species (UHF/ROHF))
    """

    # Config settings
    settings: Settings
    compute_density_cube: bool = True
    compute_electrostatic_potential_cube: bool = True
    compute_num_occupied_orbitals: NonNegativeInt = 1
    compute_num_virtual_orbitals: NonNegativeInt = 1

    # Results
    calc_uuid: UUID | None = None

    dipole: Annotated[Vector3D | None, AfterValidator(round_optional_vector3d(6))] = None
    quadrupole: Annotated[Matrix3x3 | None, AfterValidator(round_optional_matrix3x3(6))] = None

    mulliken_charges: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    lowdin_charges: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None

    wiberg_bond_orders: list[tuple[NonNegativeInt, NonNegativeInt, NonNegativeFloat]] = []
    mayer_bond_orders: list[tuple[NonNegativeInt, NonNegativeInt, NonNegativeFloat]] = []

    density_cube: PropertyCube | None = None
    density_cube_alpha: PropertyCube | None = None
    density_cube_beta: PropertyCube | None = None
    density_cube_difference: PropertyCube | None = None

    electrostatic_potential_cube: PropertyCube | None = None

    molecular_orbitals: dict[NonNegativeInt, MolecularOrbitalCube] = {}
    molecular_orbitals_alpha: dict[NonNegativeInt, MolecularOrbitalCube] = {}
    molecular_orbitals_beta: dict[NonNegativeInt, MolecularOrbitalCube] = {}



from stjames import Constraint, Mode, OptimizationSettings, Settings


def test_set_mode_auto() -> None:
    Settings()
    assert Settings().mode == Mode.RAPID


def test_opt_settings() -> None:
    settings_rapid = Settings(mode=Mode.RAPID)
    settings_meticulous = Settings(mode=Mode.METICULOUS)

    cons = [Constraint(atoms=[1, 2], constraint_type="bond")]
    settings_careful = Settings(mode=Mode.CAREFUL, opt_settings=OptimizationSettings(constraints=cons))

    rap_opt_set = settings_rapid.opt_settings
    car_opt_set = settings_careful.opt_settings
    met_opt_set = settings_meticulous.opt_settings

    assert not rap_opt_set.constraints
    assert not met_opt_set.constraints
    assert car_opt_set.constraints == cons

    assert rap_opt_set.energy_threshold == 5e-5
    assert rap_opt_set.max_gradient_threshold == 5e-3
    assert rap_opt_set.rms_gradient_threshold == 3.5e-3

    assert car_opt_set.energy_threshold == 1e-6
    assert car_opt_set.max_gradient_threshold == 9e-4
    assert car_opt_set.rms_gradient_threshold == 6e-4

    assert met_opt_set.energy_threshold == 1e-6
    assert met_opt_set.max_gradient_threshold == 3e-5
    assert met_opt_set.rms_gradient_threshold == 2e-5



import pytest

from stjames import Atom, Molecule, MoleculeReadError, PeriodicCell

valid_extxyz = """
5
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_num_atoms = """
6
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

not_digit_num_atoms = """
v
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

many_num_atoms = """
6 9
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""
no_num_atoms = """
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

xyz_style = """
5
Comment
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

missing_lattice = """
5
Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

missing_properties = """
5
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0"
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_properites = """
5
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3foo:1
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_extra = """
5
Lattice="6.0 0.0 0.0 0.0 6.0 0.0 0.0 0.0 6.0 3.14" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_equals = """
5
Lattice="6.0 0.0 =0.0 0.0 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_str = """
5
Lattice="6.0 0.0 0.0 hi 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_extra_string = """
5
Lattice="6.0 0.0 0.0 0.0 sup 6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""


incorrect_lattice_single_quote = """
5
Lattice="6.0 0.0 0.0 0.0 6.0 '0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_double_quote = """
5
Lattice="6.0 0.0 0.0 0.0 "6.0 0.0 0.0 0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_double_single_quote = """
5
Lattice="6.0 0.0 0.0 0.0 '6.0 0.0 0.0 '0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""

incorrect_lattice_double_double_quote = """
5
Lattice="6.0 0.0 "0.0 0.0 6.0 0.0 0.0 "0.0 6.0" Properties=species:S:1:pos:R:3
C        0.0        0.0        0.0
H        0.0        0.0        1.0
H        1.0        0.0        0.0
H        0.0        1.0        0.0
H        1.0        1.0        1.0
"""


expected_cell = (
    (6.0, 0.0, 0.0),
    (0.0, 6.0, 0.0),
    (0.0, 0.0, 6.0),
)

expected_atoms = [
    Atom(atomic_number=6, position=(0.0, 0.0, 0.0)),  # C
    Atom(atomic_number=1, position=(0.0, 0.0, 1.0)),  # H
    Atom(atomic_number=1, position=(1.0, 0.0, 0.0)),  # H
    Atom(atomic_number=1, position=(0.0, 1.0, 0.0)),  # H
    Atom(atomic_number=1, position=(1.0, 1.0, 1.0)),  # H
]

expected_molecule = Molecule(
    charge=0,
    multiplicity=1,
    atoms=expected_atoms,
    cell=PeriodicCell(lattice_vectors=expected_cell),
)


def test_molecule_from_extxyz_valid() -> None:
    """
    Test case for valid extxyz string.
    """
    molecule = Molecule.from_extxyz(valid_extxyz)
    assert molecule == expected_molecule, f"Valid case failed: got {molecule}, expected {expected_molecule}"


@pytest.mark.parametrize(
    "invalid_extxyz",
    [
        incorrect_num_atoms,
        no_num_atoms,
        not_digit_num_atoms,
        many_num_atoms,
        xyz_style,
        missing_lattice,
        missing_properties,
        incorrect_properites,
        incorrect_lattice_extra,
        incorrect_lattice_equals,
        incorrect_lattice_str,
        incorrect_lattice_extra_string,
        incorrect_lattice_single_quote,
        incorrect_lattice_double_quote,
        incorrect_lattice_double_single_quote,
        incorrect_lattice_double_double_quote,
    ],
)
def test_molecule_from_extxyz_invalid(invalid_extxyz: str) -> None:
    """
    Test case for invalid extxyz strings, ensuring they raise MoleculeReadError.
    """
    with pytest.raises(MoleculeReadError):
        Molecule.from_extxyz(invalid_extxyz)



from ..base import Base
from ..types import UUID
from .workflow import Workflow


class HydrogenBondAcceptorSite(Base):
    atom_idx: int  # zero-indexed
    pkbhx: float
    position: tuple[float, float, float]
    name: str | None = None


class HydrogenBondBasicityWorkflow(Workflow):
    do_csearch: bool = True
    do_optimization: bool = True

    # UUID of optimization
    optimization: UUID | None = None

    # hydrogen-bond-acceptor sites
    hba_sites: list[HydrogenBondAcceptorSite] = []  # noqa: RUF012



from typing import Annotated, Optional

from pydantic import AfterValidator

from ..base import Base, round_float, round_optional_float
from ..mode import Mode
from .workflow import DBCalculation, Workflow


class Tautomer(Base):
    energy: Annotated[float, AfterValidator(round_float(6))]
    weight: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    predicted_relative_energy: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None

    # UUIDs, optionally
    structures: list[DBCalculation] = []


class TautomerWorkflow(Workflow):
    mode: Mode = Mode.CAREFUL
    tautomers: list[Tautomer] = []



from pytest import mark

from stjames.pdb import PDB, fetch_pdb, pdb_from_mmcif_filestring, pdb_from_pdb_filestring, pdb_object_to_pdb_filestring


def test_1ema() -> None:
    """Green fluorescent protein."""
    fetch_pdb("1EMA")


def test_read_pdb_filestring() -> None:
    """Rest reading of a pdb string."""
    with open("tests/data/1ema.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    assert pdb.description.code == "1EMA"

    json = pdb.model_dump()
    PDB.model_validate(json)


def test_read_mmcif_filestring() -> None:
    """Rest reading of a mmcif string."""
    with open("tests/data/1ema.cif") as f:
        data = f.read()
    pdb = pdb_from_mmcif_filestring(data)
    assert pdb.description.code == "1EMA"

    json = pdb.model_dump()
    PDB.model_validate(json)


# fmt: off
@mark.regression
@mark.parametrize(
    "code",
    [
        # Codes from molecules of the month August 2024–January 2025
        "7S6B", "8UCS", "8UPL", "7CGO", "2ZVY", "1F4V", "6YKM", "6E10", "6E11",
        "3VCM", "2X0B", "6OS0", "1N9U", "1O86", "2V0Z", "6KI1", "6KI2", "7YYO",
        "6TJV", "7EGL", "7CYF", "7EGK", "7ZCG", "3FRT", "6AP1",
        # Codes created by o1
        "1CRN", "1MBN", "4HHB", "1HHO", "1BNA", "1CAG", "2JHO", "1EVV", "3ZOJ",
        "4AGG", "2Y69", "6R1V", "6ND2", "7NZ6", "1S72", "3G5U", "7DFT", "6AI0",
        "6NG2", "1A0I", "1B7C", "1C8R", "1D4T", "1E7O", "1F9J", "1G5K", "1H8L",
        "1I2M", "1J3N", "1K4P", "1M7R", "1N8S", "1O9T", "1P0U", "1Q1V", "1R2W",
        "1S3X", "1T4Y", "1U5Z", "1V6A", "1W7B", "1X8C", "1Y9D", "1Z0E", "2A1F",
        "2B2G", "2C3H", "2D4I", "2E5J", "2F6K", "2G7L", "2H8M", "2I9N", "2J0O",
        "2K1P", "2L2Q", "2N4S", "2O5T", "2P6U", "2Q7V", "2R8W", "2V2A", "2W3B",
        "2X4C", "2Y5D", "2Z6E", "3A7F", "3B8G", "3C9H", "3D0I", "3E1J", "3F2K",
        "3G3L", "3H4M", "3I5N", "3J6O", "3K7P", "3L8Q", "3N0S", "3O1T", "3P2U",
        "3Q3V", "3S5X", "3T6Y", "3U7Z", "3V8A", "3W9B", "3X0C", "4A3F", "4B4G",
        "4C5H", "4D6I", "4E7J", "4F8K", "4G9L", "4H0M", "4I1N", "4J2O", "4K3P",
        "4L4Q", "4M5R", "4N6S", "4O7T", "4P8U", "4Q9V", "4R0W", "4S1X",
    ]
)  # fmt: on
def test_pdb(code: str) -> None:
    pdb = fetch_pdb(code)

    json = pdb.model_dump()
    PDB.model_validate(json)

def test_from_pdb_to_pdb_2qto() -> None:
    with open("tests/data/2qto.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    filestring = pdb_object_to_pdb_filestring(pdb)
    pdb2 = pdb_from_pdb_filestring(filestring)

    assert pdb == pdb2

def test_from_pdb_to_pdb_1ema() -> None:
    with open("tests/data/1ema.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    filestring = pdb_object_to_pdb_filestring(pdb)
    pdb2 = pdb_from_pdb_filestring(filestring)

    assert pdb == pdb2



from typing import Annotated

from pydantic import AfterValidator

from ..base import round_optional_float
from ..types import UUID, FloatPerAtom, round_optional_float_per_atom
from .workflow import Workflow


class FukuiIndexWorkflow(Workflow):
    # UUID of optimization
    optimization: UUID | None = None

    global_electrophilicity_index: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    fukui_positive: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    fukui_negative: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    fukui_zero: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None



import re
from pathlib import Path
from typing import Annotated, Iterable, Optional, Self

import pydantic
from pydantic import AfterValidator, NonNegativeInt, PositiveInt, ValidationError

from .atom import Atom
from .base import Base, round_float, round_optional_float
from .periodic_cell import PeriodicCell
from .types import (
    FloatPerAtom,
    Matrix3x3,
    Vector3D,
    Vector3DPerAtom,
    round_optional_float_per_atom,
    round_optional_matrix3x3,
    round_optional_vector3d,
    round_optional_vector3d_per_atom,
    round_vector3d_per_atom,
)


class MoleculeReadError(RuntimeError):
    pass


class VibrationalMode(Base):
    frequency: Annotated[float, AfterValidator(round_float(3))]  # in cm-1
    reduced_mass: Annotated[float, AfterValidator(round_float(3))]  # amu
    force_constant: Annotated[float, AfterValidator(round_float(3))]  # mDyne/Å
    displacements: Annotated[Vector3DPerAtom, AfterValidator(round_vector3d_per_atom(6))]  # Å


class Molecule(Base):
    charge: int
    multiplicity: PositiveInt
    atoms: list[Atom]

    # for periodic boundary conditions
    cell: Optional[PeriodicCell] = None

    energy: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None  # in Hartree
    scf_iterations: Optional[NonNegativeInt] = None
    scf_completed: Optional[bool] = None
    elapsed: Annotated[Optional[float], AfterValidator(round_optional_float(3))] = None  # in seconds

    homo_lumo_gap: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None  # in eV

    gradient: Annotated[Optional[Vector3DPerAtom], AfterValidator(round_optional_vector3d_per_atom(6))] = None  # Hartree/Å
    stress: Annotated[Optional[Matrix3x3], AfterValidator(round_optional_matrix3x3(6))] = None  # Hartree/Å

    velocities: Annotated[Optional[Vector3DPerAtom], AfterValidator(round_optional_vector3d_per_atom(6))] = None  # Å/fs

    mulliken_charges: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    mulliken_spin_densities: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    dipole: Annotated[Optional[Vector3D], AfterValidator(round_optional_vector3d(6))] = None  # in Debye

    vibrational_modes: Optional[list[VibrationalMode]] = None

    zero_point_energy: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    thermal_energy_corr: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    thermal_enthalpy_corr: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    thermal_free_energy_corr: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None

    smiles: Optional[str] = None

    def __len__(self) -> int:
        return len(self.atoms)

    def distance(self, atom1: PositiveInt, atom2: PositiveInt) -> float:
        r"""
        Get the distance between atoms.

        >>> mol = Molecule.from_xyz("H 0 1 0\nH 0 0 1")
        >>> mol.distance(1, 2)
        1.4142135623730951
        """
        return sum((q2 - q1) ** 2 for q1, q2 in zip(self.atoms[atom1 - 1].position, self.atoms[atom2 - 1].position)) ** 0.5  # type: ignore [no-any-return,unused-ignore]

    @property
    def coordinates(self) -> Vector3DPerAtom:
        return [a.position for a in self.atoms]

    def translated(self, vector: Vector3D) -> Self:
        r"""
        Translate the molecule by a vector.

        >>> mol = Molecule.from_xyz("H 0 0 0\nH 0 0 1")
        >>> print(mol.translated((1, 0, 0)).to_xyz())
        2
        <BLANKLINE>
        H     1.0000000000    0.0000000000    0.0000000000
        H     1.0000000000    0.0000000000    1.0000000000
        """

        def translated(position: Vector3D) -> Vector3D:
            return tuple(q + v for q, v in zip(position, vector, strict=True))  # type: ignore [return-value]

        atoms = [atom.copy(update={"position": translated(atom.position)}) for atom in self.atoms]

        return self.copy(update={"atoms": atoms})

    @property
    def atomic_numbers(self) -> list[NonNegativeInt]:
        return [a.atomic_number for a in self.atoms]

    @property
    def sum_energy_zpe(self) -> Optional[float]:
        if (self.energy is None) or (self.zero_point_energy is None):
            return None
        return self.energy + self.zero_point_energy

    @property
    def sum_energy_thermal_corr(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_energy_corr is None):
            return None
        return self.energy + self.thermal_energy_corr

    @property
    def sum_energy_enthalpy(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_enthalpy_corr is None):
            return None
        return self.energy + self.thermal_enthalpy_corr

    @property
    def sum_energy_free_energy(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_free_energy_corr is None):
            return None
        return self.energy + self.thermal_free_energy_corr

    @pydantic.model_validator(mode="after")
    def check_electron_sanity(self) -> Self:
        num_electrons = sum(self.atomic_numbers) - self.charge
        num_unpaired_electrons = self.multiplicity - 1
        if (num_electrons - num_unpaired_electrons) % 2 != 0:
            raise ValueError(
                f"The combination of {num_electrons} electrons, charge {self.charge}, and multiplicity {self.multiplicity} is impossible. "
                "Double-check the charge and multiplicity values given and verify that they are correct."
            )

        return self

    @classmethod
    def from_file(cls: type[Self], filename: Path | str, format: str | None = None, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Read a molecule from a file.

        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile("w+", suffix=".xyz") as f:
        ...    _ = f.write("2\nComment\nH 0 0 0\nF 0 0 1")
        ...    _ = f.seek(0)
        ...    mol = Molecule.from_file(f.name)
        >>> print(mol.to_xyz())
        2
        <BLANKLINE>
        H     0.0000000000    0.0000000000    0.0000000000
        F     0.0000000000    0.0000000000    1.0000000000
        """
        filename = Path(filename)
        if not format:
            format = filename.suffix[1:]

        with open(filename) as f:
            match format:
                case "xyz":
                    return cls.from_xyz_lines(f.readlines(), charge=charge, multiplicity=multiplicity)
                case "extxyz":
                    return cls.from_extxyz_lines(f.readlines(), charge=charge, multiplicity=multiplicity)
                case _:
                    raise ValueError(f"Unsupported {format=}")

    @classmethod
    def from_xyz(cls: type[Self], xyz: str, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Generate a Molecule from an XYZ string.

        Note: only supports single molecule inputs.

        >>> len(Molecule.from_xyz("2\nComment\nH 0 0 0\nH 0 0 1"))
        2
        """
        return cls.from_xyz_lines(xyz.strip().splitlines(), charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_xyz_lines(cls: type[Self], lines: Iterable[str], charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        lines = list(lines)
        if len(lines[0].split()) == 1:
            natoms = lines[0].strip()
            if not natoms.isdigit() or (int(lines[0]) != len(lines) - 2):
                raise MoleculeReadError(f"First line of XYZ file should be the number of atoms, got: {lines[0]} != {len(lines) - 2}")
            lines = lines[2:]

        try:
            return cls(atoms=[Atom.from_xyz(line) for line in lines], charge=charge, multiplicity=multiplicity)
        except (ValueError, ValidationError) as e:
            raise MoleculeReadError("Error reading molecule from xyz") from e

    def to_xyz(self, comment: str = "", out_file: Path | str | None = None) -> str:
        r"""
        Generate an XYZ string.

        >>> mol = Molecule.from_xyz("2\nComment\nH 0 1 2\nF 1 2 3")
        >>> print(mol.to_xyz(comment="HF"))
        2
        HF
        H     0.0000000000    1.0000000000    2.0000000000
        F     1.0000000000    2.0000000000    3.0000000000
        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as directory:
        ...     file = Path(directory) / "mol.xyz"
        ...     out = mol.to_xyz(comment="HF", out_file=file)
        ...     with file.open() as f:
        ...         Molecule.from_xyz(f.read()).to_xyz("HF") == out
        True
        """
        geom = "\n".join(map(str, self.atoms))
        out = f"{len(self)}\n{comment}\n{geom}"

        if out_file:
            with Path(out_file).open("w") as f:
                f.write(out)

        return out

    @classmethod
    def from_extxyz(cls: type[Self], extxyz: str, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Generate a Molecule from a EXTXYZ string. Currently only supporting Lattice and Properties fields.

        >>> Molecule.from_extxyz('''
        ... 2
        ... Lattice="6.0 0.0 0.0 6.0 0.0 0.0 6.0 0.0 0.0"Properties=species:S:1:pos:R:3
        ... H 0 0 0
        ... H 0 0 1
        ... ''').cell.lattice_vectors
        ((6.0, 0.0, 0.0), (6.0, 0.0, 0.0), (6.0, 0.0, 0.0))
        """

        return cls.from_extxyz_lines(extxyz.strip().splitlines(), charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_extxyz_lines(cls: type[Self], lines: Iterable[str], charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        # ensure first line is number of atoms
        lines = list(lines)
        if len(lines[0].split()) == 1:
            natoms = lines[0].strip()
            if not natoms.isdigit() or (int(lines[0]) != len(lines) - 2):
                raise MoleculeReadError(f"First line of EXTXYZ file should be the number of atoms, got: {lines[0]} != {len(lines) - 2}")
            lines = lines[1:]
        else:
            raise MoleculeReadError(f"First line of EXTXYZ should be only an int denoting number of atoms. Got {lines[0].split()}")

        # ensure second line contains key-value pairs
        if "=" not in lines[0]:
            raise MoleculeReadError(f"Invalid property line, got {lines[0]}")

        cell = parse_comment_line(lines[0])
        lines = lines[1:]

        try:
            return cls(atoms=[Atom.from_xyz(line) for line in lines], cell=cell, charge=charge, multiplicity=multiplicity)
        except (ValueError, ValidationError) as e:
            raise MoleculeReadError("Error reading molecule from extxyz") from e


def parse_comment_line(line: str) -> PeriodicCell:
    """
    currently only supporting lattice and porperites fields from comment line
    modify in future to support other fields from comment from_xyz_lines
    ex: name, mulitplicity, charge, etc.
    """
    cell = None
    # Regular expression to match key="value", key='value', or key=value
    pattern = r"(\S+?=(?:\".*?\"|\'.*?\'|\S+))"
    pairs = re.findall(pattern, line)

    prop_dict = {}
    for pair in pairs:
        key, value = pair.split("=", 1)
        if key.lower() == "lattice":
            value = value.strip("'\"").split()
            if len(value) != 9:
                raise MoleculeReadError(f"Lattice should have 9 entries got {len(value)}")

            # Convert the value to a 3x3 tuple of tuples of floats
            try:
                cell = tuple(tuple(map(float, value[i : i + 3])) for i in range(0, 9, 3))
            except ValueError:
                raise MoleculeReadError(f"Lattice should be floats, got {value}")

            prop_dict[key] = value

        elif key.lower() == "properties":
            if value.lower() != "species:s:1:pos:r:3":
                raise MoleculeReadError(f"Only accepting properties of form species:S:1:pos:R:3, got {value}")
            prop_dict[key] = value
        else:
            raise MoleculeReadError(f"Currently only accepting lattice and propery keys. Got {key}")

    if cell is None:
        raise MoleculeReadError("Lattice field is required but missing.")

    if "properties" not in [key.lower() for key in prop_dict.keys()]:
        raise MoleculeReadError(f"Property field is required, got keys {prop_dict.keys()}")
    return PeriodicCell(lattice_vectors=cell)



"""Conformer Search Workflow."""

from abc import ABC
from typing import Annotated, Self, Sequence, TypeVar

from pydantic import AfterValidator, BaseModel, Field, field_validator, model_validator

from ..base import LowercaseStrEnum
from ..constraint import Constraint
from ..method import Method, XTBMethod
from ..mode import Mode
from ..types import UUID, FloatPerAtom, round_float_per_atom
from .multistage_opt import MultiStageOptMixin
from .workflow import Workflow

_sentinel = object()

_T = TypeVar("_T")
_U = TypeVar("_U")


def check_sentinel(value: _T, default: _U) -> _T | _U:
    """Return value unless _sentinel, then return default."""
    return default if value is _sentinel else value


class ScreeningSettings(BaseModel):
    """
    Settings for determining unique and useful conformers.

    :param energy_threshold: maximum relative energy for screening
    :param rotational_constants_threshold: maximum difference in rotational constants for screening
    :param rmsd: Cartesian RMSD for screening
    :param max_confs: maximum number of conformers to keep
    """

    energy_threshold: float | None = None  # kcal/mol
    rotational_constants_threshold: float | None = 0.02
    rmsd: float | None = 0.25
    max_confs: int | None = None


class ConformerGenSettings(BaseModel):
    """
    Conformer generation settings.

    Conformers are generated and an initial screening is performed to remove duplicates and high-energy conformers.

    :param mode: Mode for calculations
    :param conf_opt_method: method for the optimization
    :param screening: post-generation screening settings
    :param constraints: constraints for conformer generation
    :param nci: add a constraining potential for non-covalent interactions
    :param max_confs: maximum number of conformers to keep
    """

    mode: Mode = Mode.RAPID
    conf_opt_method: XTBMethod = Method.GFN_FF
    screening: ScreeningSettings | None = None
    constraints: Sequence[Constraint] = tuple()
    nci: bool = False
    max_confs: int | None = None

    def __str__(self) -> str:
        """Return a string representation of the ConformerGenSettings."""
        return repr(self)

    def __repr__(self) -> str:
        """Return a string representation of the ConformerGenSettings."""
        return f"<{type(self).__name__} {self.mode.name}>"


class ETKDGSettings(ConformerGenSettings):
    """
    Settings for ETKDG conformer generation.

    Inherited:
    :param mode: Mode for calculations
    :param conf_opt_method: method for the optimization
    :param screening: post-generation screening settings
    :param constraints: constraints for conformer generation
    :param nci: add a constraining potential for non-covalent interactions (not supported in ETKDG)
    :param max_confs: maximum number of conformers to keep

    New:
    :param num_initial_confs: number of initial conformers to generate
    :param num_confs_considered: number of conformers to consider for optimization
    :param num_confs_taken: number of final conformers to take
    :param max_mmff_energy: MMFF energy cutoff
    :param max_mmff_iterations: MMFF optimization iterations
    """

    num_initial_confs: int = 300
    num_confs_considered: int = 100
    max_mmff_iterations: int = 500
    max_mmff_energy: float | None = 30

    @field_validator("constraints")
    def check_constraints(cls, constraints: Sequence[Constraint]) -> Sequence[Constraint]:
        if constraints:
            raise ValueError("ETKDG does not support constraints")

        return tuple(constraints)

    @field_validator("nci")
    def check_nci(cls, nci: bool) -> bool:
        if nci:
            raise ValueError("ETKDG does not support NCI")

        return nci

    @model_validator(mode="after")
    def validate_and_build(self) -> Self:
        match self.mode:
            case Mode.MANUAL:
                pass
            case Mode.RECKLESS:
                self.num_initial_confs = 200
                self.num_confs_considered = 50
                self.max_confs = 20 if self.max_confs is None else self.max_confs
                self.max_mmff_energy = 20
            case Mode.RAPID:
                self.max_confs = 50 if self.max_confs is None else self.max_confs
                self.conf_opt_method = Method.GFN0_XTB
            case _:
                raise NotImplementedError(f"Unsupported mode: {self.mode}")

        return self


class iMTDSpeeds(LowercaseStrEnum):
    MEGAQUICK = "megaquick"
    SUPERQUICK = "superquick"
    QUICK = "quick"
    NORMAL = "normal"
    EXTENSIVE = "extensive"


class iMTDSettings(ConformerGenSettings, ABC):
    """
    Settings for iMTD style conformer generation.

    RECKLESS:
        - GFN-FF//MTD(GFN-FF)
        - Megaquick
            - No GC
            - No rotamer metadynamics
            - Energy window = 5.0
            - Run scaling factor = 0.5
            - 6 MTD runs
    RAPID:
        - GFN0//MTD(GFN-FF)
        - Superquick
            - No GC
            - No rotamer metadynamics
            - Energy window = 5.0
            - Run scaling factor = 0.5
            - 6 MTD runs
    CAREFUL:
        - GFN2//MTD(GFN-FF)
        - Quick
            - GC (for iMTD-GC)
            - Rotamer metadynamics (for iMTD-GC)
            - Energy window = 5.0
            - Run scaling factor = 0.5
            - 6 MTD runs
    METICULOUS:
        - GFN2//MTD(GFN2)
        - "Normal"
            - GC (for iMTD-GC)
            - Rotamer metadynamics (for iMTD-GC)
            - Energy window = 6.0
            - Run scaling factor = 1
            - 14 MTD runs (2 with extreme values)


    See https://github.com/crest-lab/crest/blob/5ca82feb2ec4df30a0129db957163c934f085952/src/choose_settings.f90#L202
    and https://github.com/crest-lab/crest/blob/5ca82feb2ec4df30a0129db957163c934f085952/src/confparse.f90#L825
    for how quick, superquick, and megaquick are defined.

    Additional notes:
        Extensive mode
        - GC
        - Rotamer metadynamics
        - Energy window = 8.0
        - Run scaling factor = 2
        - 14 MTD runs (2 with extreme values)

        --NCI may switch things to QUICK?

    Inherited:
    :param mode: Mode for calculations
    :param conf_opt_method: method for the optimization
    :param screening: post-generation screening settings (not used)
    :param constraints: constraints to add
    :param nci: add an ellipsoide potential around the input structure
    :param max_confs: maximum number of conformers to keep

    New:
    :param mtd_method: method for the metadynamics
    :param speed: speed of the calculations (CREST specific setting)
    :param reopt: re-optimize conformers (corrects for the lack of rotamer metadynamics and GC)
    :param free_energy_weights: calculate frequencies and re-weight based on free energies
    """

    mtd_method: XTBMethod = Method.GFN_FF
    mtd_runtype: str = "imtd-gc"

    speed: iMTDSpeeds = iMTDSpeeds.QUICK
    reopt: bool = _sentinel  # type: ignore [assignment]
    free_energy_weights: bool = False

    @model_validator(mode="after")
    def validate_and_build_imtdgc_settings(self) -> Self:
        match self.mode:
            case Mode.MANUAL:
                if self.reopt is _sentinel:
                    raise ValueError("Must specify reopt with MANUAL mode")
            case Mode.RECKLESS:  # GFN-FF//MTD(GFN-FF)
                self.max_confs = 20 if self.max_confs is None else self.max_confs
                self.speed = iMTDSpeeds.MEGAQUICK
                self.reopt = check_sentinel(self.reopt, True)
            case Mode.RAPID:  # GFN0//MTD(GFN-FF)
                self.max_confs = 50 if self.max_confs is None else self.max_confs
                self.speed = iMTDSpeeds.SUPERQUICK
                self.conf_opt_method = Method.GFN0_XTB
                self.reopt = check_sentinel(self.reopt, True)
            case Mode.CAREFUL:  # GFN2//MTD(GFN-FF)
                self.speed = iMTDSpeeds.QUICK
                self.conf_opt_method = Method.GFN2_XTB
                self.reopt = check_sentinel(self.reopt, False)
            case Mode.METICULOUS:  # GFN2//MTD(GFN2)
                self.speed = iMTDSpeeds.NORMAL
                self.mtd_method = Method.GFN2_XTB
                self.conf_opt_method = Method.GFN2_XTB
                self.reopt = check_sentinel(self.reopt, False)
            # case Mode.EXTREME: # GFN2//MTD(GFN2)
            #     self.mtd_method = Method.GFN2_XTB
            #     self.conf_opt_method = Method.GFN2_XTB
            #     self.speed = iMTDSpeeds.EXTENSIVE
            #     self.reopt = check_sentinel(self.reopt, False)
            case _:
                raise NotImplementedError(f"Unsupported mode: {self.mode}")

        return self


class iMTDGCSettings(iMTDSettings):
    run_type: str = "imtdgc"


class iMTDsMTDSettings(iMTDSettings):
    run_type: str = "imtd-smtd"


class ConformerGenMixin(BaseModel):
    """
    Mixin for workflows that need conformer generation.

    :param conf_gen_mode: Mode for calculations
    :param conf_gen_settings: settings for conformer generation
    :param constraints: constraints to add
    :param nci: add a constraining potential for non-covalent interactions
    :param max_confs: maximum number of conformers to keep
    """

    conf_gen_mode: Mode = Mode.RAPID
    conf_gen_settings: ConformerGenSettings = _sentinel  # type: ignore [assignment]
    constraints: Sequence[Constraint] = tuple()
    nci: bool = False
    max_confs: int | None = None

    @model_validator(mode="after")
    def validate_and_build_conf_gen_settings(self) -> Self:
        """Validate and build the ConformerGenSettings."""
        if self.conf_gen_settings is not _sentinel and self.conf_gen_mode != Mode.MANUAL:
            raise ValueError("Cannot specify conf_gen_settings with non-MANUAL mode")

        match self.conf_gen_mode:
            case Mode.MANUAL:
                if self.conf_gen_settings is _sentinel:
                    raise ValueError("Must specify conf_gen_settings with MANUAL mode")

            case Mode.RECKLESS | Mode.RAPID:
                # ETKDGSettings will error if constraints or nci are set
                self.conf_gen_settings = ETKDGSettings(mode=self.conf_gen_mode, constraints=self.constraints, nci=self.nci, max_confs=self.max_confs)
            case Mode.CAREFUL | Mode.METICULOUS:
                self.conf_gen_settings = iMTDSettings(mode=self.conf_gen_mode, constraints=self.constraints, nci=self.nci, max_confs=self.max_confs)

            case _:
                raise NotImplementedError(f"Unsupported mode: {self.conf_gen_mode}")

        return self


class ConformerSearchMixin(ConformerGenMixin, MultiStageOptMixin):
    """
    Mixin for workflows that need conformer search—a combination of conformer generation and optimization.

    Inherited (ConformerGenMixin):
    :param conf_gen_mode: Mode for conformer generation
    :param mso_mode: Mode for MultiStageOptSettings
    :param conf_gen_settings: settings for conformer generation
    :param nci: add a constraining potential for non-covalent interactions

    Inherited (MultiStageOptMixin):
    :param multistage_opt_settings: set by mso_mode unless mode=MANUAL (ignores additional settings if set)
    :param solvent: solvent to use
    :param xtb_preopt: pre-optimize with xtb (sets based on mode when None)
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    Inherited (Both):
    :param constraints: constraints to add (diamond inheritance, works as expected)
    """

    def __str__(self) -> str:
        """Return a string representation of the ConformerSearch workflow."""
        return repr(self)

    def __repr__(self) -> str:
        """Return a string representation of the ConformerSearch workflow."""
        return f"<{type(self).__name__} {self.conf_gen_mode.name} {self.mso_mode.name}>"

    @model_validator(mode="after")
    def remove_ts_constraints(self) -> Self:
        """
        Remove constraints from optimization if a TS.

        Also affects Manual Mode.
        """
        msos = self.multistage_opt_settings
        if msos.transition_state and msos.constraints:
            msos.constraints = []
            for opt_set in msos.optimization_settings:
                opt_set.opt_settings.constraints = []

        return self


class ConformerSearchWorkflow(ConformerSearchMixin, Workflow):
    """
    ConformerSearch Workflow.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param conf_gen_mode: Mode for calculations
    :param conf_gen_settings: settings for conformer generation
    :param mso_mode: Mode for MultiStageOptSettings
    :param multistage_opt_settings: set by mode unless mode=MANUAL (ignores additional settings if set)
    :param solvent: solvent to use
    :param xtb_preopt: pre-optimize with xtb (sets based on mode when None)
    :param constraints: constraints to add
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    Ignored:
    :param mode: Mode to use (not used)

    New:
    :param conformer_uuids: list of UUIDs of the Molecules generated
    :param energies: energies of the molecules
    """

    # Results
    conformer_uuids: list[list[UUID | None]] = Field(default_factory=list)
    energies: Annotated[FloatPerAtom, AfterValidator(round_float_per_atom(6))] = Field(default_factory=list)



from typing import Annotated

import numpy as np
from numpy.typing import NDArray
from pydantic import AfterValidator

from ..base import Base, round_optional_float
from ..molecule import Molecule
from ..settings import Settings
from ..types import UUID
from .workflow import Workflow


class ScanPoint(Base):
    index: int
    molecule: Molecule
    energy: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    uuid: UUID | None = None


class ScanSettings(Base):
    type: str  # "bond", "angle", "dihedral" - make Enum later
    atoms: list[int]
    start: float
    stop: float
    num: int

    def vals(self) -> NDArray[np.float64]:
        return np.linspace(self.start, self.stop, self.num)  # type: ignore [return-value, unused-ignore]

    class Config:
        from_attributes = True


class ScanWorkflow(Workflow):
    scan_settings: ScanSettings
    calc_settings: Settings
    calc_engine: str

    # UUIDs of scan points
    scan_points: list[UUID | None] = []



"""Contains various file handling helper functions."""

import builtins
import gzip
from typing import Any

from requests import get  # type: ignore [import-untyped]

from .mmcif import mmcif_dict_to_data_dict, mmcif_string_to_mmcif_dict
from .pdb import pdb_dict_to_data_dict, pdb_string_to_pdb_dict


def open(path: str, *args, **kwargs) -> Any:  # type: ignore [no-untyped-def]
    """Opens a file at a given path, works out what filetype it is, and parses
    it accordingly.

    For example:
    open('/path/to/file.pdb', data_dict=True)

    This will parse file.pdb as a .pdb file, but only go as far as converting it
    to an atomium data dictionary.

    If the file extension is .gz, the file will be unzipped first.

    :param str path: the location of the file.
    :param bool file_dict: if ``True``, parsing will stop at the file ``dict``.
    :param bool data_dict: if ``True``, parsing will stop at the data ``dict``.
    :rtype: ``File``"""

    if str(path)[-3:] == ".gz":
        try:
            with gzip.open(path) as f:
                filestring = f.read().decode()
        except Exception:
            with gzip.open(path, "rt") as f:
                filestring = f.read()
        return parse_string(filestring, path[:-3], *args, **kwargs)
    else:
        try:
            with builtins.open(path) as f:
                filestring = f.read()
        except Exception:
            with builtins.open(path, "rb") as f:
                filestring = f.read()  # type: ignore [assignment]
        return parse_string(filestring, path, *args, **kwargs)


def fetch(code: str, *args, **kwargs) -> Any:  # type: ignore [no-untyped-def]
    """Fetches a file from a remote location via HTTP.

    If a PDB code is given, the .cif form of that struture will be fetched from
    the RCSB servers. If that code is given an extension, that file format will
    be obtained instead of .cif. If a URL is given, the function will simply
    look in that location.

    For example:
    fetch('1lol.mmtf', file_dict=True)

    This will get the .mmtf version of structure 1LOL, but only go as far as
    converting it to an atomium file dictionary.

    :param str code: the file to fetch.
    :param bool file_dict: if ``True``, parsing will stop at the file ``dict``.
    :param bool data_dict: if ``True``, parsing will stop at the data ``dict``.
    :raises ValueError: if no file is found.
    :rtype: ``File``"""

    if code.startswith("http"):
        url = code
    elif code.endswith(".mmtf"):
        url = "https://mmtf.rcsb.org/v1.0/full/{}".format(code[:-5].lower())
    else:
        if "." not in code:
            code += ".pdb"
        url = "https://files.rcsb.org/view/" + code.lower()
    response = get(url, stream=True)
    if response.status_code == 200:
        text = response.content if code.endswith(".mmtf") else response.text
        return parse_string(text, code, *args, **kwargs)
    raise ValueError("Could not find anything at {}".format(url))


def parse_string(filestring: str, path: str, file_dict: bool = False, data_dict: bool = False) -> Any:
    """Takes a filestring and parses it in the appropriate way. You must provide
    the string to parse itself, and some other string that ends in either .cif,
    .mmtf, or .cif - that will determine how the file is parsed.

    (If this cannot be inferred from the path string, atomium will guess based
    on the filestring contents.)

    :param str filestring: the contents of some file.
    :param str path: the filename of the file of origin.
    :param bool file_dict: if ``True``, parsing will stop at the file ``dict``.
    :param bool data_dict: if ``True``, parsing will stop at the data ``dict``.
    :rtype: ``File``"""

    file_func, data_func = get_parse_functions(filestring, path)
    parsed = file_func(filestring)
    if not file_dict:
        parsed = data_func(parsed)
    return parsed


def get_parse_functions(filestring: str, path: str) -> Any:
    """Works out which parsing functions to use for a given filestring and
    returns them.

    (If this cannot be inferred from the path string, atomium will guess based
    on the filestring contents.)

    :param str filestring: the filestring to inspect.
    :param str path: the path to inspect.
    :rtype: ``tuple``"""

    if "." in path:
        ending = path.split(".")[-1]
        if ending in ("mmtf", "cif", "pdb"):
            return {
                "cif": (mmcif_string_to_mmcif_dict, mmcif_dict_to_data_dict),
                "pdb": (pdb_string_to_pdb_dict, pdb_dict_to_data_dict),
            }[ending]
    elif "_atom_sites" in filestring:
        return (mmcif_string_to_mmcif_dict, mmcif_dict_to_data_dict)
    else:
        return (pdb_string_to_pdb_dict, pdb_dict_to_data_dict)
